
# q-expansion of newform 400.4.n.c, downloaded from the LMFDB on 05 May 2024.

# We generate the q-expansion using the Hecke eigenvalues a_p at the primes.
# Each a_p is given as a linear combination
# of the following basis for the coefficient ring.

def make_data():

    from sage.all import prod, floor, prime_powers, gcd, QQ, primes_first_n, next_prime, RR

    def discrete_log(elts, gens, mod):
        # algorithm 2.2, page 16 of https://arxiv.org/abs/0903.2785
        def table_gens(gens, mod):
            T = [1]
            n = len(gens)
            r = [None]*n
            s = [None]*n
            for i in range(n):
                beta = gens[i]
                r[i] = 1
                N = len(T)
                while beta not in T:
                    for Tj in T[:N]:
                        T.append((beta*Tj) % mod)
                    beta = (beta*gens[i]) % mod
                    r[i] += 1
                s[i] = T.index(beta)
            return T, r, s
        T, r, s = table_gens(gens, mod)
        n = len(gens)
        N = [ prod(r[:j]) for j in range(n) ]
        Z = lambda s: [ (floor(s/N[j]) % r[j]) for j in range(n)]
        return [Z(T.index(elt % mod)) for elt in elts]
    def extend_multiplicatively(an):
        for pp in prime_powers(len(an)-1):
            for k in range(1, (len(an) - 1)//pp + 1):
                if gcd(k, pp) == 1:
                    an[pp*k] = an[pp]*an[k]
    from sage.all import PolynomialRing, NumberField, ZZ
    R = PolynomialRing(QQ, "x")
    f = R(poly_data)
    K = NumberField(f, "a")
    betas = [K([c/ZZ(den) for c in num]) for num, den in basis_data]
    convert_elt_to_field = lambda elt: sum(c*beta for c, beta in zip(elt, betas))
    # convert aps to K elements
    primes = primes_first_n(len(aps_data))
    good_primes = [p for p in primes if not p.divides(level)]
    aps = map(convert_elt_to_field, aps_data)
    if not hecke_ring_character_values:
        # trivial character
        char_values = dict(zip(good_primes, [1]*len(good_primes)))
    else:
        gens = [elt[0] for elt in hecke_ring_character_values]
        gens_values = [convert_elt_to_field(elt[1]) for elt in hecke_ring_character_values]
        char_values = dict([(
            p,prod(g**k for g, k in zip(gens_values, elt)))
            for p, elt in zip(good_primes, discrete_log(good_primes, gens, level))
            ])
    an_list_bound = next_prime(primes[-1])
    an = [0]*an_list_bound
    an[1] = 1
    
    from sage.all import PowerSeriesRing
    PS = PowerSeriesRing(K, "q")
    for p, ap in zip(primes, aps):
        if p.divides(level):
            euler_factor = [1, -ap]
        else:
            euler_factor = [1, -ap, p**(weight - 1) * char_values[p]]
        k = RR(an_list_bound).log(p).floor() + 1
        foo = (1/PS(euler_factor)).padded_list(k)
        for i in range(1, k):
            an[p**i] = foo[i]
    extend_multiplicatively(an)
    return PS(an)
level = 400
weight = 4
poly_data = [81, 0, -17, 0, 1]

# The entries in the following list give a basis for the
# coefficient ring in terms of a root of the defining polynomial above.
# Each line consists of the coefficients of the numerator, and a denominator.
basis_data  = [[[1, 0, 0, 0], 1], [[0, -8, 0, 1], 9], [[-153, 26, 18, -1], 9], [[153, 26, -18, -1], 9]]

hecke_ring_character_values = [[351, [-1, 0, 0, 0]], [101, [1, 0, 0, 0]], [177, [0, 1, 0, 0]]]
aps_data = [[0, 0, 0, 0], [0, 0, -1, 0], [0, 0, 0, 0], [0, 0, 0, 1], [0, 0, -5, 5], [15, -15, 0, 0], [-65, -65, 0, 0], [0, 0, 10, 10], [0, 0, -17, 0], [0, 4, 0, 0], [0, 0, -15, 15], [115, 115, 0, 0], [18, 0, 0, 0], [0, 0, 31, 0], [0, 0, 0, -47], [-265, 265, 0, 0], [0, 0, -10, -10], [-722, 0, 0, 0], [0, 0, 0, 5], [0, 0, 5, -5], [135, -135, 0, 0], [0, 0, -80, -80], [0, 0, 7, 0], [0, 104, 0, 0], [615, 615, 0, 0], [942, 0, 0, 0], [0, 0, -145, 0], [0, 0, 0, 93], [0, -1124, 0, 0], [55, -55, 0, 0], [0, 0, 0, -87], [0, 0, -185, 185], [-1145, -1145, 0, 0], [0, 0, -90, -90], [0, 756, 0, 0], [0, 0, 85, -85], [2035, 2035, 0, 0], [0, 0, 63, 0], [0, 0, 0, -183], [895, -895, 0, 0], [0, 0, -310, -310], [3102, 0, 0, 0], [0, 0, 105, -105], [2015, -2015, 0, 0], [155, 155, 0, 0], [0, 0, -180, -180], [0, 0, -265, 265], [0, 0, -441, 0], [0, 0, 0, -499], [0, -1236, 0, 0], [1215, -1215, 0, 0], [0, 0, 280, 280], [3338, 0, 0, 0], [0, 0, -245, 245], [-2505, -2505, 0, 0], [0, 0, 439, 0], [0, 3756, 0, 0], [0, 0, -575, 575], [1035, 1035, 0, 0], [-2782, 0, 0, 0], [0, 0, 295, 0], [5255, -5255, 0, 0], [0, 0, 0, 1189], [0, 0, 445, -445], [-3225, 3225, 0, 0], [-4325, -4325, 0, 0], [0, 0, 355, -355], [-6625, -6625, 0, 0], [0, 0, 0, -11], [0, -556, 0, 0], [575, -575, 0, 0], [0, 0, -820, -820], [0, 0, 0, -231], [-8625, 8625, 0, 0], [0, 0, 150, 150], [0, 0, 1351, 0], [0, -11644, 0, 0], [4275, 4275, 0, 0], [-11358, 0, 0, 0], [0, 7216, 0, 0], [0, 0, 1210, 1210], [12278, 0, 0, 0], [0, 0, -735, 735], [7335, -7335, 0, 0], [0, 0, -620, -620], [0, 0, -153, 0], [0, -9464, 0, 0], [575, 575, 0, 0], [-3578, 0, 0, 0], [0, 0, 783, 0], [0, 0, 0, 1629], [0, 0, 1240, 1240], [0, 0, 0, -2303], [0, 0, -125, 125], [0, 0, 1490, 1490], [0, 0, -913, 0], [0, 15604, 0, 0], [1322, 0, 0, 0], [0, 0, -2649, 0], [-16618, 0, 0, 0], [0, 0, 0, 2829], [-16085, -16085, 0, 0], [0, 0, 2415, 0], [0, -20864, 0, 0], [0, 0, -445, 445], [15175, 15175, 0, 0], [0, 0, 0, -235], [6495, -6495, 0, 0], [0, 0, -940, -940], [-16062, 0, 0, 0], [0, 0, 0, 2145], [12495, -12495, 0, 0], [-19145, -19145, 0, 0], [0, 0, 670, 670], [0, 0, 2365, -2365], [6618, 0, 0, 0], [0, 0, 887, 0], [0, 0, 0, 1665], [17255, -17255, 0, 0], [0, 0, -350, -350], [4358, 0, 0, 0], [455, -455, 0, 0], [2875, 2875, 0, 0], [0, 0, -801, 0], [0, 0, -1145, 1145], [-9202, 0, 0, 0], [0, 284, 0, 0], [0, 0, -1880, -1880], [0, 0, 0, 1697], [4975, -4975, 0, 0], [0, 0, -2750, -2750], [0, 0, -585, 0], [0, 0, -255, 255], [2555, 2555, 0, 0], [-7878, 0, 0, 0], [0, 24464, 0, 0], [-3985, 3985, 0, 0], [0, 0, 0, 677], [-26805, -26805, 0, 0], [0, -11816, 0, 0], [0, 0, -3165, 3165], [-42602, 0, 0, 0], [0, 0, -3905, 0], [0, 0, 0, -715], [0, -4, 0, 0], [0, 0, 1780, 1780], [-16785, 16785, 0, 0], [18775, 18775, 0, 0], [0, 0, -1130, -1130], [0, 0, 1423, 0], [-10045, -10045, 0, 0], [19818, 0, 0, 0], [0, 0, -433, 0], [0, 0, 0, -2031], [0, 0, 0, -4819], [0, 0, 905, -905], [0, 0, -2460, -2460], [0, -19464, 0, 0], [-25465, -25465, 0, 0], [-7578, 0, 0, 0], [0, 0, 0, -2203], [28055, -28055, 0, 0], [0, 0, 0, -4343], [0, 0, -1125, 1125], [-11305, -11305, 0, 0], [0, 0, 2535, 0], [0, 0, 2385, -2385], [-42925, -42925, 0, 0]]
